import ctypes

from picowrap import constants
from picowrap.picoscope import Picoscope


class Channel:
    """Channel
    Picoscope instance

    Attributes:
        picoscope: Picoscope,
        name: str,
        voltage: str,
        coupling: str = "DC",
        offset: float = 0,

    """

    def __init__(
        self,
        picoscope: Picoscope,
        name: str,
        voltage: str,
        offset: float = 0,
        coupling: str = "DC",
    ) -> None:

        self.picoscope = picoscope

        self.number = constants.CHANNEL_COR[name]
        self.voltage = picoscope.VOLTAGE_COR[voltage.upper()]
        self.offset = offset
        self.coupling = constants.COUPLING_COR[coupling]

        self._create()

        self.maxsamples = sum(self.picoscope.trigscope)
        self.buffer_max = (ctypes.c_int16 * self.maxsamples)()
        self.buffer_min = (ctypes.c_int16 * self.maxsamples)()
        self.set_data_buffers()

    def _create(self):
        self.picoscope.set_channel(
            self.number,
            1,
            self.coupling,
            self.voltage,
            self.offset,
        )

    def trigger(self, threshold: float, direction: str, delay: int = 0, wait: int = 0):
        """
        threshold (float): the value in V at which the trigger will fire

        direction (str): the direction in which the signal must move to cause a
        trigger. The following directions are supported: ABOVE, BELOW,
        RISING, FALLING and RISING_OR_FALLING.

        delay (nb_samples): the time between the trigger occurring and the first sample.

        wait (ms): the number of milliseconds the device will wait
        if no trigger occurs. If this is set to zero,
        the scope device will wait indefinitely for a trigger.
        """

        adc_max = self.picoscope.maximum_value()
        adc_threshold = round(
            (threshold * adc_max) / constants.VOLTAGE_CON[self.voltage]
        )
        self.picoscope.set_simple_trigger(
            1,
            self.number,
            adc_threshold,
            constants.DIRECTION_COR[direction.upper()],
            delay,
            wait,
        )

    def set_data_buffers(self):
        self.picoscope.set_data_buffers(
            self.number, self.buffer_max, self.buffer_min, self.maxsamples
        )

    def get_values(self):
        nb_samples_in_buff = self.picoscope.get_values(0, self.maxsamples)
        if nb_samples_in_buff != self.maxsamples:
            print("nb_samples_in_buff != self.maxsamples")
            print(f"{nb_samples_in_buff} != {self.maxsamples}")

        max_value = self.picoscope.maximum_value()

        value = [
            x * constants.VOLTAGE_CON[self.voltage] / max_value for x in self.buffer_max
        ]
        time_interval_ns, _ = self.picoscope.get_time_base_2(
            self.picoscope.timebase, sum(self.picoscope.trigscope)
        )
        if time_interval_ns * 10**-9 != self.picoscope.sampling:
            print(
                f"WARNING sampling ({self.picoscope.sampling})",
                f"!= real time_interval ({time_interval_ns * 10**-9}) ",
                f"Error = {self.picoscope.sampling-time_interval_ns * 10**-9} [s]",
            )

        return value



"""A base module interacting with the picoscope sdk.

For using this CBase you should have PicoSDK installed.
Check https://www.picotech.com/downloads

  Typical usage example:

  class Picoscope(Cpicoscope):

"""

import ctypes
import math
import os
from typing import Any

from picowrap import constants, errors


class CBase:
    """CPicoscope class for interacting with PicoSDK in python

    Attributes:
        status: A dict for storing sdk response,
        chandle: A c_int16 for storing handle,
        name: A str for accessing model relative dll and fonction in sdk
        clib_path: A str representing the dll path

        PICO_INFO: A dict for converting hex to str info
        PICO_STATUS: A dict for converting hex to str status
        PICO_STATUS_LOOKUP: Inverted dict of PICO_STATUS

    TODO: Docstring (from 118)
    """

    PICO_INFO = constants.PICO_INFO
    PICO_STATUS = constants.PICO_STATUS
    PICO_STATUS_LOOKUP = {v: k for k, v in PICO_STATUS.items()}
    VOLTAGE_COR = None  # Define in subclass
    # MAX_MEMORY = 2**29 = 536 MS (or better 512 MS)

    def __init__(self, name: str, sampling: float) -> None:
        """Inits CPicoscope"""

        self.status: dict = {}
        self.chandle = ctypes.c_int16()
        self.name = name
        self.sampling = sampling

        self.clib_path = self.find_clib()

    @property
    def timebase(self):
        raise NotImplementedError("To be define in subclass")

    def find_clib(self) -> str:
        """Find PicoSDK in PATH.

        Iterate through PATH directories to find the specific DLL relative to the entered path

        Raises:
            errors.CannotFindPicoSDKError: PicoSDK not found in PATH.

        Returns:
            str: fullpath of the DLL
        """
        for directory in os.environ["PATH"].split(os.pathsep):
            fname = os.path.join(directory, self.name) + ".dll"
            if os.path.isfile(fname):
                return fname
        raise errors.CannotFindPicoSDKError(
            f"PicoSDK {self.name} not found, check PATH"
        )

    @property
    def _clib(self) -> ctypes.WinDLL:
        """Load DLL

        Raises:
            errors.CannotOpenPicoSDKError: Opening DLL fail

        Returns:
           ctypes.WinDLL: DLL object
        """
        try:
            result = ctypes.WinDLL(self.clib_path)
        except OSError as err:
            raise errors.CannotOpenPicoSDKError(
                f"PicoSDK {self.name} not compatible (check 32 vs 64-bit): {err}"
            )
        return result

    def _assert_pico_ok(self, status: int) -> None:
        """Checks for PICO_OK status return else return specified Error

        Args:
            status (int): Status hex value

        Raises:
            errors.PicoSDKCtypesError: Response not ok
        """
        if status != self.PICO_STATUS["PICO_OK"]:
            raise errors.PicoSDKCtypesError(
                f"PicoSDK returned {self.PICO_STATUS_LOOKUP[status]}"
            )

    def _c_func(self, c_name: str, return_type, argument_types: list) -> Any:
        """Create a python c function

        Args:
            c_name (str): Function name
            return_type (c_uint32): Return type of the function
            argument_types (list[c_type]): List of c_type used in argument

        Returns:
            _type_: _description_
        """
        c_function = getattr(self._clib, c_name)
        c_function.restype = return_type
        c_function.argtypes = argument_types
        return c_function

    def open_unit(self, serial=None, resolution=None) -> None:
        """

        int16_t *handle,
        int8_t  *serial

        ### Raises:

        """
        c_name = self.name + "OpenUnit"
        return_type = ctypes.c_uint32

        serial = None  # Serial not implemented

        if resolution:
            argument_types = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_int32]
            c_function = self._c_func(c_name, return_type, argument_types)
            resolution = constants.DEVICE_RES[resolution]
            cresolution = ctypes.c_int32()
            cresolution.value = resolution
            self.status["openunit"] = c_function(
                ctypes.byref(self.chandle), serial, cresolution
            )
        else:
            argument_types = [ctypes.c_void_p, ctypes.c_char_p]
            c_function = self._c_func(c_name, return_type, argument_types)
            self.status["openunit"] = c_function(ctypes.byref(self.chandle), serial)
        try:
            self._assert_pico_ok(self.status["openunit"])

        except errors.PicoSDKCtypesError as typerr:

            if self.PICO_STATUS_LOOKUP[self.status["openunit"]] == "PICO_NOT_FOUND":
                raise Exception(f"Pico {self.name} not connected") from typerr

            powerstate = self.status["openunit"]

            if powerstate == 282:  # PICO_POWER_SUPPLY_NOT_CONNECTED
                self.status["ChangePowerSource"] = self._change_power_source(282)

            elif powerstate == 286:  # PICO_USB3_0_DEVICE_NON_USB3_0_PORT
                self.status["ChangePowerSource"] = self._change_power_source(286)
            else:
                raise Exception(
                    "[DEV] Status from openUnit not implemented"
                ) from typerr

    def set_data_buffers(
        self, channel, buff_max, buff_min, buff_len
    ) -> tuple[list[int], list[int]]:
        """

        # Handle = Chandle
        # source = ps3000A_channel_A = 0
        # Buffer max = ctypes.byref(bufferAMax)
        # Buffer min = ctypes.byref(bufferAMin)
        # Buffer length = maxsamples
        # Segment index = 0
        # Ratio mode = ps3000A_Ratio_Mode_None = 0

        int16_t            handle,
        int32_t            channel Or Port,
        int16_t           *bufferMax,
        int16_t           *bufferMin,
        int32_t            bufferLth,
        uint32_t           segmentIndex,
        PS3000a_RATIO_MODE mode
        """
        c_name = self.name + "SetDataBuffers"
        return_type = ctypes.c_uint32
        argument_types = [
            ctypes.c_int16,
            ctypes.c_int32,
            ctypes.c_void_p,
            ctypes.c_void_p,
            ctypes.c_int32,
            ctypes.c_uint32,
            ctypes.c_int32,
        ]

        seg_index = 0
        ratio_mode = 0
        c_function = self._c_func(c_name, return_type, argument_types)
        self.status["SetDataBuffers"] = c_function(
            self.chandle,
            channel,
            ctypes.byref(buff_max),
            ctypes.byref(buff_min),
            buff_len,
            seg_index,
            ratio_mode,
        )
        self._assert_pico_ok(self.status["SetDataBuffers"])
        return buff_max, buff_min

    def is_ready(self):
        """
        int16_t  handle,
        int16_t *ready
        """
        c_name = self.name + "IsReady"
        return_type = ctypes.c_uint32
        argument_types = [ctypes.c_int16, ctypes.c_void_p]

        c_function = self._c_func(c_name, return_type, argument_types)

        ready = ctypes.c_int16(0)
        self.status["isReady"] = c_function(self.chandle, ctypes.byref(ready))
        return ready

    def run_block(self, pre_trig, post_trig, timebase, *args):
        """
        int16_t            handle,
        int32_t            noOfPreTriggerSamples,
        int32_t            noOfPostTriggerSamples,
        uint32_t           timebase,
        int16_t            oversample,
        int32_t           *timeIndisposedMs,
        uint32_t           segmentIndex,
        ps3000aBlockReady  lpReady,
        void              *pParameter
        """
        if not args:
            oversample = 0
            time_indisposed = None
            segment_index = 0
            lp_read = None
            p_parameter = None
        else:
            oversample, time_indisposed, segment_index, lp_read, p_parameter = args

        c_name = self.name + "RunBlock"
        return_type = ctypes.c_uint32
        argument_types = [
            ctypes.c_int16,
            ctypes.c_int32,
            ctypes.c_int32,
            ctypes.c_uint32,
            ctypes.c_int16,
            ctypes.c_void_p,
            ctypes.c_uint32,
            ctypes.c_void_p,
            ctypes.c_void_p,
        ]

        c_function = self._c_func(c_name, return_type, argument_types)
        self.status["runblock"] = c_function(
            self.chandle,
            pre_trig,
            post_trig,
            timebase,
            oversample,
            time_indisposed,
            segment_index,
            lp_read,
            p_parameter,
        )
        self._assert_pico_ok(self.status["runblock"])

    def _change_power_source(self, powerstate: int):
        """
        Change power source
        """
        c_name = self.name + "ChangePowerSource"
        return_type = ctypes.c_uint32
        argument_types = [ctypes.c_int16, ctypes.c_uint32]

        c_function = self._c_func(c_name, return_type, argument_types)
        self.status["ChangePowerSource"] = c_function(self.chandle, powerstate)
        self._assert_pico_ok(self.status["ChangePowerSource"])
        return self.status["ChangePowerSource"]

    def set_channel(self, channel, enabled, coupling, voltage, offset):
        """
        channel = PS3000A_CHANNEL_A = 0
        enabled = 1
        coupling type = PS3000A_DC = 1
        range = PS3000A_10V = 8
        analogue offset = 0 V

        PS3000a_CHANNEL  channel,
        int16_t          enabled,
        PS3000a_COUPLING type,
        PS3000a_RANGE    range,
        float            analogOffset
        """
        c_name = self.name + "SetChannel"
        return_type = ctypes.c_uint32
        argument_types = [
            ctypes.c_int16,
            ctypes.c_int32,
            ctypes.c_int16,
            ctypes.c_int32,
            ctypes.c_int32,
            ctypes.c_float,
        ]

        c_function = self._c_func(c_name, return_type, argument_types)
        self.status[f"setCh{channel}"] = c_function(
            self.chandle,
            channel,
            enabled,
            coupling,
            voltage,
            offset,
        )
        self._assert_pico_ok(self.status[f"setCh{channel}"])

    def set_simple_trigger(
        self, enable, source, threshold, direction, delay, autotrigger_ms
    ):
        """
        int16_t                      handle,
        int16_t                      enable,
        PS3000A_CHANNEL              source,
        int16_t                      threshold,
        PS3000A_THRESHOLD_DIRECTION  direction,
        uint32_t                     delay,
        int16_t                      autoTrigger_ms
        Enable = 0
        Source = ps3000A_channel_A = 0
        Threshold = 1024 ADC counts
        Direction = ps3000A_Falling = 3
        Delay = 0
        autoTrigger_ms = 0
        """
        c_name = self.name + "SetSimpleTrigger"
        return_type = ctypes.c_uint32
        argument_types = [
            ctypes.c_int16,
            ctypes.c_int16,
            ctypes.c_int32,
            ctypes.c_int16,
            ctypes.c_int32,
            ctypes.c_uint32,
            ctypes.c_int16,
        ]

        c_function = self._c_func(c_name, return_type, argument_types)
        self.status["trigger"] = c_function(
            self.chandle,
            enable,
            source,
            threshold,
            direction,
            delay,
            autotrigger_ms,
        )
        self._assert_pico_ok(self.status["trigger"])

    def get_time_base_2(
        self, timebase: int, n_samples: int, oversample=1, segment_index=0
    ):
        """
        handle, device identifier returned by ps3000aOpenUnit
        timebase, see timebase guide. This value can be supplied to
        ps3000aRunBlock to define the sampling interval.
        noSamples, the number of samples required
        segmentIndex, the index of the memory segment to use

        * timeIntervalNanoseconds, on exit, the time interval between
        readings at the selected timebase. Use NULL if not required.
        oversample, not used
        * maxSamples, on exit, the maximum number of samples available.
        The scope allocates a certain amount of memory for internal
        overheads and this may vary depending on the number of segments,
        number of channels enabled, and the timebase chosen. Use NULL if
        not required.
        """
        c_name = self.name + "GetTimebase2"
        return_type = ctypes.c_uint32
        argument_types = [
            ctypes.c_int16,
            ctypes.c_uint32,
            ctypes.c_int32,
            ctypes.c_void_p,
            ctypes.c_int16,
            ctypes.c_void_p,
            ctypes.c_uint32,
        ]

        c_function = self._c_func(c_name, return_type, argument_types)

        time_interval_ns = ctypes.c_float()
        returned_max_samples = ctypes.c_int16()

        self.status["GetTimebase"] = c_function(
            self.chandle,
            timebase,
            n_samples,
            ctypes.byref(time_interval_ns),
            oversample,
            ctypes.byref(returned_max_samples),
            segment_index,
        )
        self._assert_pico_ok(self.status["GetTimebase"])
        return time_interval_ns.value, returned_max_samples.value

    def get_values(self, start_index, nb_samples, segment_index=0):
        """
        handle, device identifier returned by ps3000aOpenUnit
        startIndex, a zero-based index that indicates the start point for
        data collection. It is measured in sample intervals from the start of
        the buffer.
        * noOfSamples, on entry, the number of samples required. On exit,
        the actual number retrieved. The number of samples retrieved will
        not be more than the number requested, and the data retrieved
        starts at startIndex.
        downSampleRatio, the downsampling factor that will be applied to
        the raw data
        downSampleRatioMode, which downsampling mode to use. The
        available values are: -
        PS3000A_RATIO_MODE_NONE (downSampleRatio is ignored)
        PS3000A_RATIO_MODE_AGGREGATE
        PS3000A_RATIO_MODE_AVERAGE
        PS3000A_RATIO_MODE_DECIMATE
        AGGREGATE, AVERAGE, DECIMATE are single-bit constants that can be
        ORed to apply multiple downsampling modes to the same data
        segmentIndex, the zero-based number of the memory segment
        where the data is stored
        * overflow, on exit, a set of flags that indicate whether an
        overvoltage has occurred on
        # Handle = chandle
        # start index = 0
        # noOfSamples = ctypes.byref(cmaxSamples)
        # DownSampleRatio = 0
        # DownSampleRatioMode = 0
        # SegmentIndex = 0
        # Overflow = ctypes.byref(overflow)
        """
        c_name = self.name + "GetValues"
        return_type = ctypes.c_uint32
        argument_types = [
            ctypes.c_int16,
            ctypes.c_uint32,
            ctypes.c_void_p,
            ctypes.c_uint32,
            ctypes.c_int32,
            ctypes.c_uint32,
            ctypes.c_void_p,
        ]

        c_function = self._c_func(c_name, return_type, argument_types)

        down_sample_ratio = 0
        down_sample_ratio_mode = 0

        # overflow = (ctypes.c_int16 * 10)()
        overflow_input = ctypes.c_void_p(0)
        overflow = overflow_input

        self.status["GetValues"] = c_function(
            self.chandle,
            start_index,
            ctypes.byref(ctypes.c_int32(nb_samples)),
            down_sample_ratio,
            down_sample_ratio_mode,
            segment_index,
            ctypes.byref(overflow),
        )
        self._assert_pico_ok(self.status["GetValues"])

        if overflow_input != overflow:
            print(overflow.value)
        samples = nb_samples
        return samples

    def maximum_value(self) -> int:
        """
        int16_t  handle,
        int16_t *value
        """
        c_name = self.name + "MaximumValue"
        return_type = ctypes.c_uint32
        argument_types = [
            ctypes.c_int16,
            ctypes.c_void_p,
        ]

        c_function = self._c_func(c_name, return_type, argument_types)

        max_adc = ctypes.c_int16()

        self.status["maximumValue"] = c_function(self.chandle, ctypes.byref(max_adc))
        self._assert_pico_ok(self.status["maximumValue"])
        return max_adc.value

    def stop(self):
        c_name = self.name + "Stop"
        return_type = ctypes.c_uint32
        argument_types = [
            ctypes.c_int16,
        ]

        c_function = self._c_func(c_name, return_type, argument_types)

        self.status["stop"] = c_function(self.chandle)
        self._assert_pico_ok(self.status["stop"])

    def close_unit(self):
        c_name = self.name + "CloseUnit"
        return_type = ctypes.c_uint32
        argument_types = [ctypes.c_int16]

        c_function = self._c_func(c_name, return_type, argument_types)

        self.status["close"] = c_function(self.chandle)
        self._assert_pico_ok(self.status["close"])


class C3000a(CBase):
    VOLTAGE_COR = constants.VOLTAGE_COR["ps3000a"]

    @property
    def timebase(self):
        if self.sampling in [1e-9, 2e-9, 4e-9]:
            return int(math.log2(self.sampling * 1000000000))
        if (self.sampling * 1e9) % 8 == 0:
            return int(self.sampling * 125000000 + 2)
        raise Exception("Sampling should be %8[ns]")

    def run_block(self, pre_trig, post_trig, timebase, *args):
        """
        int16_t            handle,
        int32_t            noOfPreTriggerSamples,
        int32_t            noOfPostTriggerSamples,
        uint32_t           timebase,
        # int16_t            oversample,
        int32_t           *timeIndisposedMs,
        uint32_t           segmentIndex,
        ps3000aBlockReady  lpReady,
        void              *pParameter
        """
        if not args:
            oversample = 0
            time_indisposed = None
            segment_index = 0
            lp_read = None
            p_parameter = None
        else:
            oversample, segment_index, lp_read, p_parameter = args

        c_name = self.name + "RunBlock"
        return_type = ctypes.c_uint32
        argument_types = [
            ctypes.c_int16,
            ctypes.c_int32,
            ctypes.c_int32,
            ctypes.c_uint32,
            ctypes.c_int16,
            ctypes.c_void_p,
            ctypes.c_uint32,
            ctypes.c_void_p,
            ctypes.c_void_p,
        ]

        c_function = self._c_func(c_name, return_type, argument_types)
        self.status["runblock"] = c_function(
            self.chandle,
            pre_trig,
            post_trig,
            timebase,
            oversample,
            time_indisposed,
            segment_index,
            lp_read,
            p_parameter,
        )

        self._assert_pico_ok(self.status["runblock"])

    def get_time_base_2(
        self, timebase: int, n_samples: int, oversample=1, segment_index=0
    ):
        """
        handle, device identifier returned by ps3000aOpenUnit
        timebase, see timebase guide. This value can be supplied to
        ps3000aRunBlock to define the sampling interval.
        noSamples, the number of samples required
        segmentIndex, the index of the memory segment to use

        * timeIntervalNanoseconds, on exit, the time interval between
        readings at the selected timebase. Use NULL if not required.
        oversample, not used
        * maxSamples, on exit, the maximum number of samples available.
        The scope allocates a certain amount of memory for internal
        overheads and this may vary depending on the number of segments,
        number of channels enabled, and the timebase chosen. Use NULL if
        not required.
        """
        c_name = self.name + "GetTimebase2"
        return_type = ctypes.c_uint32
        argument_types = [
            ctypes.c_int16,
            ctypes.c_uint32,
            ctypes.c_int32,
            ctypes.c_void_p,
            ctypes.c_int16,
            ctypes.c_void_p,
            ctypes.c_uint32,
        ]

        c_function = self._c_func(c_name, return_type, argument_types)

        time_interval_ns = ctypes.c_float()
        returned_max_samples = ctypes.c_int16()

        self.status["GetTimebase"] = c_function(
            self.chandle,
            timebase,
            n_samples,
            ctypes.byref(time_interval_ns),
            oversample,
            ctypes.byref(returned_max_samples),
            segment_index,
        )
        self._assert_pico_ok(self.status["GetTimebase"])
        return time_interval_ns.value, returned_max_samples.value


class C5000a(CBase):
    VOLTAGE_COR = constants.VOLTAGE_COR["ps5000a"]

    @property
    def timebase(self):
        resolution = 8
        if resolution == 8:
            if self.sampling in [1e-9, 2e-9, 4e-9]:
                return int(math.log2(self.sampling * 1000000000))
            if (self.sampling * 1e9) % 8 == 0:
                return int(self.sampling * 125000000 + 2)
            raise Exception("Sampling should be %8[ns]")
        if resolution == 12:
            if self.sampling in [2e-9, 4e-9, 8e-9]:
                return int(math.log2(self.sampling * 1000000000))
            if (self.sampling * 1e9) % 16 == 0:
                return int(self.sampling * 62500000 + 3)
            raise Exception("Sampling should be %16[ns]")

        if resolution in [14, 15]:
            if self.sampling in [8e-9]:
                return 3
            if (self.sampling * 1e9) % 8 == 0:
                return int(self.sampling * 125000000 + 2)
            raise Exception("Sampling should be %8[ns]")

        if resolution == 16:
            if self.sampling in [16e-9]:
                return 4
            if (self.sampling * 1e9) % 16 == 0:
                return int(self.sampling * 62500000 + 3)
            raise Exception("Sampling should be %8[ns]")

    def run_block(self, pre_trig, post_trig, timebase, *args):
        """
        int16_t            handle,
        int32_t            noOfPreTriggerSamples,
        int32_t            noOfPostTriggerSamples,
        uint32_t           timebase,
        # int16_t            oversample,
        int32_t           *timeIndisposedMs,
        uint32_t           segmentIndex,
        ps3000aBlockReady  lpReady,
        void              *pParameter
        """
        if not args:
            oversample = 0
            # time_indisposed = None
            segment_index = 0
            lp_read = None
            p_parameter = None
        else:
            oversample, segment_index, lp_read, p_parameter = args

        c_name = self.name + "RunBlock"
        return_type = ctypes.c_uint32
        argument_types = [
            ctypes.c_int16,
            ctypes.c_int32,
            ctypes.c_int32,
            ctypes.c_uint32,
            # ctypes.c_int16,
            ctypes.c_void_p,
            ctypes.c_uint32,
            ctypes.c_void_p,
            ctypes.c_void_p,
        ]

        c_function = self._c_func(c_name, return_type, argument_types)
        self.status["runblock"] = c_function(
            self.chandle,
            pre_trig,
            post_trig,
            timebase,
            oversample,
            # time_indisposed,
            segment_index,
            lp_read,
            p_parameter,
        )

        self._assert_pico_ok(self.status["runblock"])

    def get_time_base_2(
        self, timebase: int, n_samples: int, oversample=1, segment_index=0
    ):
        """
        handle, device identifier returned by ps3000aOpenUnit
        timebase, see timebase guide. This value can be supplied to
        ps3000aRunBlock to define the sampling interval.
        noSamples, the number of samples required
        segmentIndex, the index of the memory segment to use

        * timeIntervalNanoseconds, on exit, the time interval between
        readings at the selected timebase. Use NULL if not required.
        oversample, not used
        * maxSamples, on exit, the maximum number of samples available.
        The scope allocates a certain amount of memory for internal
        overheads and this may vary depending on the number of segments,
        number of channels enabled, and the timebase chosen. Use NULL if
        not required.
        """
        c_name = self.name + "GetTimebase2"
        return_type = ctypes.c_uint32
        argument_types = [
            ctypes.c_int16,
            ctypes.c_uint32,
            ctypes.c_int32,
            ctypes.c_void_p,
            # ctypes.c_int16,
            ctypes.c_void_p,
            ctypes.c_uint32,
        ]

        c_function = self._c_func(c_name, return_type, argument_types)

        time_interval_ns = ctypes.c_float()
        returned_max_samples = ctypes.c_int16()

        self.status["GetTimebase"] = c_function(
            self.chandle,
            timebase,
            n_samples,
            ctypes.byref(time_interval_ns),
            # oversample,
            ctypes.byref(returned_max_samples),
            segment_index,
        )
        self._assert_pico_ok(self.status["GetTimebase"])
        return time_interval_ns.value, returned_max_samples.value


class CPicoscope(CBase):
    def __new__(cls, name: str, *args, **kwargs):
        _ = args, kwargs
        mixin = {"ps3000a": C3000a, "ps5000a": C5000a}[name]
        cls_type = type(mixin.__name__, (cls, mixin), {})
        return super().__new__(cls_type)



import matplotlib.pyplot as plt

from picowrap import Channel, Picoscope

TIME_INTERVAL = 16e-9  # [s]
DURATION = 0.5e-3  # [s]

# with Picoscope("ps3000a", TIME_INTERVAL, DURATION) as pico:
with Picoscope("ps5000a", TIME_INTERVAL, DURATION, resolution=12) as pico:

    chan_A = Channel(pico, "A", "10V")
    # chan_B = Channel(pico, "B", "10V")
    # chan_C = Channel(pico, "C", "10V")
    # chan_D = Channel(pico, "D", "10V")

    chan_A.trigger(4, "ABOVE", 0, 0)

    pico.run()

    dataA = chan_A.get_values()
    # dataB = chan_B.get_values()
    # dataC = chan_C.get_values()
    # dataD = chan_D.get_values()

time = [round(i * TIME_INTERVAL, 10) for i in range(len(dataA))]
# enumerate is 0.04[s] slower in this case

plt.plot(time, dataA)
# plt.plot(time, dataB)
# plt.plot(time, dataC)
# plt.plot(time, dataD)
plt.xlabel("Time (s)")
plt.ylabel("Voltage (V)")
plt.show()



""" Errors """


class PicoError(Exception):
    """All errors thrown explicitly by this package will be PicoError's."""


class FeatureNotSupportedError(PicoError):
    """Raised when a feature is not supported on the connected device."""


class CannotFindPicoSDKError(PicoError, IOError):
    """SDK can't be found"""


class CannotOpenPicoSDKError(PicoError, IOError):
    """SDK can't be opened"""


class DeviceNotFoundError(PicoError, IOError):
    pass


class ArgumentOutOfRangeError(PicoError, ValueError):
    pass


class ValidRangeEnumValueNotValidForThisDevice(FeatureNotSupportedError, ValueError):
    pass


class DeviceCannotSegmentMemoryError(FeatureNotSupportedError, TypeError):
    pass


class InvalidMemorySegmentsError(PicoError, ValueError):
    pass


class InvalidTimebaseError(PicoError, ValueError):
    pass


class InvalidTriggerParameters(PicoError, ValueError):
    pass


class InvalidCaptureParameters(PicoError, ValueError):
    pass


class PicoSDKCtypesError(PicoError, IOError):
    pass


class ClosedDeviceError(PicoError, IOError):
    pass


class NoChannelsEnabledError(PicoError, ValueError):
    pass


class NoValidTimebaseForOptionsError(PicoError, ValueError):
    pass


class UnknownConstantError(PicoError, TypeError):
    pass



import time
from multiprocessing import Manager, Process

import matplotlib.pyplot as plt

from picowrap import Channel, Picoscope

TIME_INTERVAL = 0.001  # [s]
DURATION = 1  # [s]


def measure(return_data_dict):
    with Picoscope("ps5000a", TIME_INTERVAL, DURATION, resolution=12) as pico:

        chan_a = Channel(pico, "A", "10V")
        chan_a.trigger(1, "ABOVE")

        pico.run()
        chan_data = chan_a.get_values()

        time_data = [round(i * TIME_INTERVAL, 10) for i in range(len(chan_data))]
        # enumerate is 0.04[s] slower in this case

        return_data_dict["Time"] = time_data
        return_data_dict["Data"] = chan_data


def count_second():
    t_0 = time.time()
    for i in range(11):
        t_1 = time.time()
        print(f"{i} : {t_1-t_0:4f}")
        time.sleep(1)


if __name__ == "__main__":

    manager = Manager()
    return_dict = manager.dict()

    job0 = Process(target=measure, args=((return_dict),))
    job1 = Process(target=count_second)

    job0.start()
    job1.start()

    job0.join()

    print("Job0: Measure is Done")
    Time = return_dict["Time"]
    Data = return_dict["Data"]
    plt.plot(Time, Data)
    plt.xlabel("Time (s)")
    plt.ylabel("Voltage (V)")
    plt.show()

    job1.join()



import ctypes

from picowrap.cbase import CPicoscope


class Picoscope(CPicoscope):
    """Picoscope class for interact with picoscope

    Attributes:
        sampling: [s] Time between 2 mesures i.e: 0.0001[s]
        duration: [s] Total time of the mesure i.e: 1[s]

    TODO : warn for max memorie
    TODO : clean adress storage
    """

    # MAX_MEMORY = 2**29 = 536 MS (512 MS)

    def __init__(
        self, name: str, sampling: float, duration: float, resolution=None
    ) -> None:
        # super(Picoscope, self).__init__(name=name)
        super().__init__(name=name, sampling=sampling)
        self.name = name

        mid_sample = int(duration / (2 * sampling))
        self.trigscope = [mid_sample, mid_sample]
        self.sampling = sampling
        self.resolution = resolution

    def __enter__(self):
        self.open_unit(resolution=self.resolution)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close_unit()

    def run(self):
        self.run_block(self.trigscope[0], self.trigscope[1], self.timebase)
        self.wait_ready()

    def wait_ready(self):
        ready = ctypes.c_int16(0)
        check = ctypes.c_int16(0)
        while ready.value == check.value:
            ready = self.is_ready()



import threading
import time

import matplotlib.pyplot as plt

from picowrap import Channel, Picoscope

TIME_INTERVAL = 0.001  # [s]
DURATION = 1  # [s]

CHANNEL_DATA = []


def measure():

    with Picoscope("ps5000a", TIME_INTERVAL, DURATION, resolution=12) as pico:

        chan_a = Channel(pico, "A", "10V")
        chan_a.trigger(1, "ABOVE")

        pico.run()

        global CHANNEL_DATA  # pylint: disable=global-statement
        # Should be use in a class context do not use global in your code

        CHANNEL_DATA = chan_a.get_values()


def count_second():
    t_0 = time.time()
    for i in range(11):
        t_1 = time.time()
        print(f"{i} : {t_1-t_0:4f}")
        time.sleep(1)


thread_daq = threading.Thread(target=measure, daemon=True)
thread_daq.start()

count_second()
thread_daq.join()

time_data = [round(i * TIME_INTERVAL, 10) for i in range(len(CHANNEL_DATA))]
# enumerate is 0.04[s] slower in this case

plt.plot(time_data, CHANNEL_DATA)
plt.xlabel("Time (s)")
plt.ylabel("Voltage (V)")
plt.show()
